
BEGIN {
  unless ($ENV{AUTHOR_TESTING}) {
    print qq{1..0 # SKIP these tests are for testing by the author\n};
    exit
  }
}

use strict;
use warnings;

# this test was generated with Dist::Zilla::Plugin::Test::NoTabs 0.15

use Test::More 0.88;
use Test::NoTabs;

my @files = (
    'bin/createrelease.pl',
    'lib/Dist/Zilla/Plugin/GitHub/CreateRelease.pm',
    't/00-load.t',
    't/author-critic.t',
    't/author-eof.t',
    't/author-eol.t',
    't/author-no-tabs.t',
    't/author-pod-coverage.t',
    't/author-pod-spell.t',
    't/author-pod-syntax.t',
    't/extract-changes.t',
    't/release-trailing-space.t',
    't/repo-name-from-url.t',
    't/repo-owner-from-url.t'
);

notabs_ok($_) foreach @files;
done_testing;
